package com.zfabrik.components;

import java.io.File;
import java.io.IOException;
import java.util.Collection;
import java.util.Objects;
import java.util.Optional;
import java.util.Properties;

import com.zfabrik.resources.IResourceHandle;
import com.zfabrik.resources.TypeRef;
import com.zfabrik.util.expression.X;

/**
 * Convenience wrapper around {@link IComponentsLookup} and {@link IComponentsManager}
 * 
 * This class provides static access methods to most used methods in {@link IComponentsLookup} and 
 * {@link IComponentsManager}.
 */
public class Components {

	/**
	 * Get a component facet by lookup through 
	 * 
	 * <pre><code>
	 * IComponentsLookup.INSTANCE.lookup(...)
	 * </code></pre>
	 * 
	 * optionally requiring support of the facet.
	 * 
	 * @param <T> Generic type of the facet
	 * @param componentName Name of the component
	 * @param clz Class of the generic type
	 * @param required if <code>true</code> the lookup must result in a non-null result or an {@link IllegalStateException} will be thrown
	 * @return The facet as implemented by the component
	 */
	public static <T> T getComponentAs(String componentName, Class<T> clz, boolean required) {
		T t = IComponentsLookup.INSTANCE.lookup(componentName, clz);
		if (t==null && required) {
			throw createUnsupportedFacetException(componentName, clz);
		}
		return t;
	}

	/**
	 * Get a component facet by lookup through 
	 * 
	 * <pre><code>
	 * IComponentsLookup.INSTANCE.lookup(...)
	 * </code></pre>
	 * 
	 * optionally requiring support of the facet.
	 * 
	 * @param <T> Generic type of the facet
	 * @param componentName Name of the component
	 * @param typeRef {@link TypeRef} of the generic type
	 * @param required if <code>true</code> the lookup must result in a non-null result or an {@link IllegalStateException} will be thrown
	 * @return The facet as implemented by the component
	 */
	public static <T> T getComponentAs(String componentName, TypeRef<T> typeRef, boolean required) {
		T t = IComponentsLookup.INSTANCE.lookup(componentName, typeRef);
		if (t==null && required) {
			throw createUnsupportedFacetException(componentName, typeRef);
		}
		return t;
	}

	/**
	 * Get a component facet by lookup through 
	 * 
	 * <pre><code>
	 * getComponentAs(componentName, clz, false)
	 * </code></pre>
	 * 
	 * i.e. with a non-required result.
	 * 
	 * @param <T> Generic type of the facet
	 * @param componentName Name of the component
	 * @param clz Class of the generic type
	 * @return The facet as implemented by the component
	 */
	public static <T> T getComponentAs(String componentName, Class<T> clz) {
		return getComponentAs(componentName, clz, false);
	}

	/**
	 * Get a component facet by lookup through 
	 * 
	 * <pre><code>
	 * getComponentAs(componentName, clz, false)
	 * </code></pre>
	 * 
	 * i.e. with a non-required result.
	 * 
	 * @param <T> Generic type of the facet
	 * @param componentName Name of the component
	 * @param typeRef {@link TypeRef} of the generic type
	 * @return The facet as implemented by the component
	 */
	public static <T> T getComponentAs(String componentName, TypeRef<T> typeRef) {
		return getComponentAs(componentName, typeRef, false);
	}

	/**
	 * Get the {@link IResourceHandle} facet of the named component. Note: This lookup is 
	 * generically supported by the system and always returns a non-null result.
	 * 
	 * @param componentName Name of the component
	 * @return The component's resource handle
	 */
	public static IResourceHandle componentHandle(String componentName) {
		return getComponentAs(componentName, IResourceHandle.class, false);
	}

	/**
	 * Add a dependency by component name and retrieve a facet of the dependency.
	 * 
	 * @param <T> Generic type of the facet
	 * @param dependant Handle of the component that will receive the dependency
	 * @param name Name of the dependency component that will be added as a dependency
	 * @param clz Class of the generic type
	 * @param required if <code>true</code> the lookup via the dependency must result in a non-null result or an {@link IllegalStateException} will be thrown
	 * @return The facet as implemented by the dependency component
	 * @return
	 */
	public static <T> T getDependencyComponentAs(IResourceHandle dependant, String name, Class<T> clz, boolean required) {
		dependant.addDependency(componentHandle(name));
		return getComponentAs(name, clz, true);
	}
	
	/**
	 * Add a dependency by component name and retrieve a facet of the dependency.
	 * 
	 * @param <T> Generic type of the facet
	 * @param dependant Handle of the component that will receive the dependency
	 * @param name Name of the dependency component that will be added as a dependency
	 * @param typeRef {@link TypeRef} of the generic type
	 * @param required if <code>true</code> the lookup via the dependency must result in a non-null result or an {@link IllegalStateException} will be thrown
	 * @return The facet as implemented by the dependency component
	 * @return
	 */
	public static <T> T getDependencyComponentAs(IResourceHandle dependant, String name, TypeRef<T> typeRef, boolean required) {
		dependant.addDependency(componentHandle(name));
		return getComponentAs(name, typeRef, true);
	}

	/**
	 * Get the component descriptor properties for the component supplied by {@link IResourceHandle}.
	 * 
	 * @param component Resource handle of the component to retrieve properties for
	 * @return Properties of the supplied component
	 */
	public static Properties getComponentProperties(IResourceHandle component) {
		return Optional.ofNullable(component.as(IComponentDescriptor.class))
				.map(IComponentDescriptor::getProperties)
				.orElseThrow(()->createUnsupportedFacetException(component.getResourceInfo().getName(), IComponentDescriptor.class));
	}

	/**
	 * Get the component descriptor properties for the component supplied by name.
	 * 
	 * @param componentName Name of the component to retrieve properties for
	 * @return Properties of the supplied component
	 */
	public static Properties getComponentProperties(String componentName) {
		return getComponentAs(componentName, IComponentDescriptor.class, true).getProperties();
	}
	
	/**
	 * Get the resource folder of a component, if any. See also {@link IComponentsManager#retrieve(String)}
	 * 
	 * @param componentName Name of the component
	 * @return Resource folder as {@link File}, or <code>null</code>, if the component has no resource folder
	 */
	public static File retrieve(String componentName) throws IOException {
		return IComponentsManager.INSTANCE.retrieve(componentName);
	}
			
	/**
	 * Find all components by query expression on their properties.
	 * 
	 * @param propertyExpression Search query for components. See {@link IComponentsManager#findComponents(X)}
	 * @return Query result
	 */
	public static Collection<String> findComponents(X propertyExpression) throws IOException {
		return IComponentsManager.INSTANCE.findComponents(propertyExpression);
	}
			
	// 
	private static <T> IllegalStateException createUnsupportedFacetException(String name, Class<T> clz) {
		return new IllegalStateException("Component "+name+" not found or not supporting facet type "+clz.getName());
	}

	private static <T> IllegalStateException createUnsupportedFacetException(String name, TypeRef<T> typeRef) {
		return new IllegalStateException("Component "+name+" not found or not supporting facet type "+Objects.toString(typeRef));
	}

}
